#ifndef ISCSI_H_
#define ISCSI_H_

/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi.h,v 1.40.8.4 2003/10/29 06:52:51 krishmnc Exp $ 
 *
 * iscsi.h
 *
 *    include for iSCSI kernel module
 * 
 */
#include <linux/list.h>
#include "iscsiAuthClient.h"
#include "iscsi-common.h"
#include "iscsi-limits.h"
#include "iscsi-task.h"
#include "iscsi-session.h"
#ifndef MIN
# define MIN(x, y)               ((x) < (y)) ? (x) : (y)
#endif

#ifndef MAX
# define MAX(x, y)    max(x,y)
#endif

struct iscsi_hba {
	struct iscsi_hba *next;
	struct Scsi_Host *host;
	unsigned int host_no;
	unsigned long flags;
	spinlock_t session_lock;
	struct list_head session_head;
	atomic_t num_sessions;
	kmem_cache_t *task_cache;
};

/* HBA flags */
#define ISCSI_HBA_ACTIVE          0
#define ISCSI_HBA_SHUTTING_DOWN   1
#define ISCSI_HBA_RELEASING       2

/* driver entry points needed by the probing code */
int iscsi_queue(struct iscsi_session * session, Scsi_Cmnd * sc,
		void (*done) (Scsi_Cmnd *));
int iscsi_squash_cmnd(struct iscsi_session * session, Scsi_Cmnd * sc);

/* flags we set on Scsi_Cmnds */
#define COMMAND_TIMEDOUT 1

/* run-time controllable logging */
#define ISCSI_LOG_ERR   1
#define ISCSI_LOG_SENSE 2
#define ISCSI_LOG_INIT  3
#define ISCSI_LOG_QUEUE 4
#define ISCSI_LOG_ALLOC 5
#define ISCSI_LOG_EH    6
#define ISCSI_LOG_FLOW  7
#define ISCSI_LOG_RETRY 8
#define ISCSI_LOG_LOGIN 9
#define ISCSI_LOG_TIMEOUT 10

#define LOG_SET(flag) (1U << (flag))
#define LOG_ENABLED(flag) (iscsi_log_settings & (1U << (flag)))

extern volatile unsigned int iscsi_log_settings;

#ifdef DEBUG
/* compile in all the debug messages and tracing */
# define INCLUDE_DEBUG_ERROR  1
# define INCLUDE_DEBUG_TRACE  1
# define INCLUDE_DEBUG_INIT   1
# define INCLUDE_DEBUG_QUEUE  1
# define INCLUDE_DEBUG_FLOW   1
# define INCLUDE_DEBUG_ALLOC  1
# define INCLUDE_DEBUG_EH     1
# define INCLUDE_DEBUG_RETRY  1
# define INCLUDE_DEBUG_TIMEOUT 1
#else
/* leave out the tracing and most of the debug messages */
# define INCLUDE_DEBUG_ERROR  1
# define INCLUDE_DEBUG_TRACE  0
# define INCLUDE_DEBUG_INIT   0
# define INCLUDE_DEBUG_QUEUE  0
# define INCLUDE_DEBUG_FLOW   0
# define INCLUDE_DEBUG_ALLOC  0
# define INCLUDE_DEBUG_EH     0
# define INCLUDE_DEBUG_RETRY  0
# define INCLUDE_DEBUG_TIMEOUT 0
#endif

#if INCLUDE_DEBUG_INIT
#  define DEBUG_INIT(fmt, args...) do { if (LOG_ENABLED(ISCSI_LOG_INIT)) printk(fmt , ## args); } while (0)
#else
#  define DEBUG_INIT(fmt, args...) do { } while (0)
#endif

#if INCLUDE_DEBUG_QUEUE
#  define DEBUG_QUEUE(fmt, args...) do { if (LOG_ENABLED(ISCSI_LOG_QUEUE)) printk(fmt , ## args); } while (0)
#else
#  define DEBUG_QUEUE(fmt, args...) do { } while (0)
#endif

#if INCLUDE_DEBUG_FLOW
#  define DEBUG_FLOW(fmt, args...) do { if (LOG_ENABLED(ISCSI_LOG_FLOW)) printk(fmt , ## args); } while (0)
#else
#  define DEBUG_FLOW(fmt, args...) do { } while (0)
#endif

#if INCLUDE_DEBUG_ALLOC
#  define DEBUG_ALLOC(fmt, args...) do { if (LOG_ENABLED(ISCSI_LOG_ALLOC)) printk(fmt , ## args); } while (0)
#else
#  define DEBUG_ALLOC(fmt, args...) do { } while (0)
#endif

#if INCLUDE_DEBUG_RETRY
#  define DEBUG_RETRY(fmt, args...) do { if (LOG_ENABLED(ISCSI_LOG_RETRY)) printk(fmt , ## args); } while (0)
#else
#  define DEBUG_RETRY(fmt, args...) do { } while (0)
#endif

#if INCLUDE_DEBUG_EH
#  define DEBUG_EH(fmt, args...) do { if (LOG_ENABLED(ISCSI_LOG_EH)) printk(fmt , ## args); } while (0)
#else
#  define DEBUG_EH(fmt, args...) do { } while (0)
#endif

#if INCLUDE_DEBUG_ERROR
#  define DEBUG_ERR(fmt, args...) do { if (LOG_ENABLED(ISCSI_LOG_ERR)) printk(fmt , ## args); } while (0)
#else
#  define DEBUG_ERR(fmt, args...) do { } while (0)
#endif

#if INCLUDE_DEBUG_TIMEOUT
#  define DEBUG_TIMEOUT(fmt, args...) do { if (LOG_ENABLED(ISCSI_LOG_TIMEOUT)) printk(fmt , ## args); } while (0)
#else
#  define DEBUG_TIMEOUT(fmt, args...) do { } while (0)
#endif

/* the Scsi_Cmnd's request_bufflen doesn't always match the actual amount of data
 * to be read or written.  Try to compensate by decoding the cdb.
 */
extern unsigned int iscsi_expected_data_length(Scsi_Cmnd * sc);

/* extract parts of the sense data from an (unsigned char *) to the beginning of sense data */
#define SENSE_KEY(sensebuf) ((sensebuf)[2] & 0x0F)
#define ASC(sensebuf)       ((sensebuf)[12])
#define ASCQ(sensebuf)       ((sensebuf)[13])

#endif
