/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2001 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      Kenneth W. Preslan <kpreslan@sistina.com>
 *
 */
#ifndef _NOLOCK_H
#define _NOLOCK_H


#define NOLOCK_LVB_SIZE (32)

#define HASH_SHIFT 8
#define HASH_SIZE (1 << HASH_SHIFT)
#define HASH_MASK (HASH_SIZE - 1)

struct nolock_space {
	spinlock_t		lock;
	struct list_head	hash[HASH_SIZE];
};

struct nolock_lock {
	struct list_head	list;
	unsigned int		hash;
	lm_lockname_t		name;
	atomic_t		count;

	unsigned int		state;

	atomic_t		lvb_count;
	lm_lvb_t		lvb;
	char			lvb_data[NOLOCK_LVB_SIZE];
	char			lvb_data_net[NOLOCK_LVB_SIZE];

	struct nolock_space	*ls;
};


/*  Macros  */

#define NOLOCK_ASSERT(x)						\
do {									\
	if (!(x)) {							\
		printk("\nnolock:  "					\
			"Assertion failed on line %d of file %s\n",	\
			__LINE__, __FILE__);				\
		BUG();							\
	}								\
} while (0)


#endif  /*  NOLOCK_DOT_H  */
