#include <stdio.h>

#define WIDTH 320
#define HEIGHT 480
#define PACKET_SIZE 384
#define HEADER_SIZE 0x8

static inline void
spca50x_move_420_block(int yTL, int yTR, int yBL, int yBR, int u, int v, 
	int rowPixels, unsigned char * rgb, int bits)
{
	const int rvScale = 91881;
	const int guScale = -22553;
	const int gvScale = -46801;
	const int buScale = 116129;
	const int yScale  = 65536;
	int r, g, b;

	g = guScale * u + gvScale * v;
	if (force_rgb) {
		r = buScale * u;
		b = rvScale * v;
	} else {
		r = rvScale * v;
		b = buScale * u;
	}

	yTL *= yScale; yTR *= yScale;
	yBL *= yScale; yBR *= yScale;

	if (bits == 24) {
		/* Write out top two pixels */
		rgb[0] = LIMIT(b+yTL); rgb[1] = LIMIT(g+yTL); rgb[2] = LIMIT(r+yTL);
		rgb[3] = LIMIT(b+yTR); rgb[4] = LIMIT(g+yTR); rgb[5] = LIMIT(r+yTR);

		/* Skip down to next line to write out bottom two pixels */
		rgb += 3 * rowPixels;
		rgb[0] = LIMIT(b+yBL); rgb[1] = LIMIT(g+yBL); rgb[2] = LIMIT(r+yBL);
		rgb[3] = LIMIT(b+yBR); rgb[4] = LIMIT(g+yBR); rgb[5] = LIMIT(r+yBR);
	} else if (bits == 16) {
		/* Write out top two pixels */
		rgb[0] = ((LIMIT(b+yTL) >> 3) & 0x1F) | ((LIMIT(g+yTL) << 3) & 0xE0);
		rgb[1] = ((LIMIT(g+yTL) >> 5) & 0x07) | (LIMIT(r+yTL) & 0xF8);

		rgb[2] = ((LIMIT(b+yTR) >> 3) & 0x1F) | ((LIMIT(g+yTR) << 3) & 0xE0);
		rgb[3] = ((LIMIT(g+yTR) >> 5) & 0x07) | (LIMIT(r+yTR) & 0xF8);

		/* Skip down to next line to write out bottom two pixels */
		rgb += 2 * rowPixels;

		rgb[0] = ((LIMIT(b+yBL) >> 3) & 0x1F) | ((LIMIT(g+yBL) << 3) & 0xE0);
		rgb[1] = ((LIMIT(g+yBL) >> 5) & 0x07) | (LIMIT(r+yBL) & 0xF8);

		rgb[2] = ((LIMIT(b+yBR) >> 3) & 0x1F) | ((LIMIT(g+yBR) << 3) & 0xE0);
		rgb[3] = ((LIMIT(g+yBR) >> 5) & 0x07) | (LIMIT(r+yBR) & 0xF8);
	}
}

int main(int argc,char * argv[])
{
  int I;
  FILE *file=fopen(argv[1],"r");
  char buffer[512];
  char data[WIDTH*HEIGHT*3];
  int stripe=0;
  int packet=0;
  int rgb=0;
  int pixel=0;
  int bytes=0;
  
  fread(buffer,0x0F+5,1,file);
  fread(buffer,HEADER_SIZE,1,file);
  if(buffer[1]!=0x3) fprintf(stderr,"Failed to read header byte\n");
  else fprintf(stderr,"Read format bytes : ");
  for(I=0;I<HEADER_SIZE;I++)
    {
      fprintf(stderr,"%02hhx ",buffer[I]);
    }
  fprintf(stderr,"\n");
  packet=HEADER_SIZE;
  
  while(!feof(file))
    {
      fread(buffer,1,1,file);
      bytes++;
      //      fprintf(stderr,"%02hhx ",buffer[0]);
      if(!(packet++%PACKET_SIZE)) 
        {
          fprintf(stderr,"Packet %d\n",buffer[0]<0?buffer[0]+256:buffer[0]);
          if(!buffer[0]) 
            {
              fprintf(stderr,"new frame at 0x%x \n",bytes);
            }
          continue;
        }

      if(rgb==WIDTH*HEIGHT*3) 
        {
          fprintf(stderr,"Read all data\n");
          break;
        }

      if(pixel<WIDTH*2)
        {
          data[rgb++]=buffer[0];
          data[rgb++]=buffer[0];
          data[rgb++]=buffer[0];
        }
      pixel++;
      if(pixel==3*WIDTH) 
        {
          pixel=0;
          fprintf(stderr,"Finished stripe %d\n",stripe++);
        }
      
    }
  if(feof(file)) fprintf(stderr,"EOF\n");
  fprintf(stdout,"P6\n%d %d\n255\n",WIDTH,HEIGHT);
  fwrite(data,WIDTH*HEIGHT*3,1,stdout);
  fflush(stdout);
  exit(0);
}


