/* ------------------------------------------------------------------------- */
/* i2c-detect.c generic I2C detection for NVIDIA display adapters	     */
/* ------------------------------------------------------------------------- */
/*   Copyright (C) 2000 Ferenc Bakonyi <fero@drama.obuda.kando.hu>
 *   Based on various sources by Simon G. Vogl
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *   
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		     */
/* ------------------------------------------------------------------------- */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/i2c.h>

/*----------------------------------------------------------------------
 *	Generic I2C device detection
 *	
 *	This fake device tries to identify all chips connected
 *	to the Riva I2C bus.
 *----------------------------------------------------------------------*/

static int unknown_probe(struct i2c_adapter *adap);
static int unknown_detach(struct i2c_client *client);
static int unknown_command(struct i2c_client *client, 
		unsigned int cmd, void *arg);

/* Addresses to scan */
static unsigned short test_normal_i2c[] = {I2C_CLIENT_END};
static unsigned short test_normal_i2c_range[] = { 0x00, 0x7f, 
						  I2C_CLIENT_END };
static unsigned short test_probe[2] = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short test_probe_range[2] = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short test_ignore[2] = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short test_ignore_range[2] = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short test_force[2] = { I2C_CLIENT_END, I2C_CLIENT_END };

static struct i2c_client_address_data test_addr_data = {
	test_normal_i2c, test_normal_i2c_range,
	test_probe, test_probe_range,
	test_ignore, test_ignore_range,
	test_force
};

/*  For registering the I2C 'driver' driver */
static struct i2c_driver i2c_driver_unknown = {
	name:		"probe",
	id:		I2C_DRIVERID_EXP0,
	flags:		I2C_DF_NOTIFY,
	attach_adapter:	unknown_probe,
	detach_client:	unknown_detach,
	command:	unknown_command,
};


/* Called when a new I2C device found */
static int unknown_attach(struct i2c_adapter *adap, int addr, unsigned short flags, int kind)		
{
	printk("i2c-detect: I2C capable chip found on %s at address %#x\n", adap->name, addr);
	return 0;
}

/* Called when a new I2C bus found */
static int unknown_probe(struct i2c_adapter *adap)
{
	/* Only attach these chips to the RIVA bus for now */
	
	if (adap->id == (I2C_ALGO_BIT | I2C_HW_B_RIVA))
	{
		return i2c_probe(adap, &test_addr_data, unknown_attach);
	}
	return 0;
}

/* Called on exit */
static int unknown_detach(struct i2c_client *client)
{
	i2c_detach_client(client);
	return 0;
}

static int unknown_command(struct i2c_client *client,
			     unsigned int cmd, void *arg)
{
	return -EINVAL;
}



/*---------------------------------------------------------------------- 
 *	Modularization
 *----------------------------------------------------------------------*/

MODULE_AUTHOR("Ferenc Bakonyi <fero@drama.obuda.kando.hu>");
MODULE_DESCRIPTION("I2C chip detection driver for NVIDIA cards");
MODULE_LICENSE("GPL");

int __init init_unknown(void) 
{
	return i2c_add_driver(&i2c_driver_unknown);
}

void __exit cleanup_unknown(void) 
{
	i2c_del_driver(&i2c_driver_unknown);
}

module_init(init_unknown);
module_exit(cleanup_unknown);
