/***************************************************************************
 * Video post-processing module for the W996[87]CF driver for Linux.       *
 *                                                                         *
 * Copyright (C) 2002 2003 by Luca Risolia <luca.risolia@studio.unibo.it>  *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify    *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation; either version 2 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with this program; if not, write to the Free Software             *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               *
 ***************************************************************************/

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>

#include "w9968cf_decoder.h"
#include "w9968cf_genericvpp.h"


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 19)
#	error "Kernel 2.4.19 is the minimum for this driver."
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	EXPORT_NO_SYMBOLS;
#endif

MODULE_AUTHOR("Luca Risolia <luca.risolia@studio.unibo.it>");
MODULE_DESCRIPTION("Video post-processing module for the "
                   "W996[87]CF JPEG USB Dual Mode Camera Chip driver");
MODULE_SUPPORTED_DEVICE("Video");
MODULE_LICENSE("GPL");


/****************************************************************************
 * Module init and exit                                                     *
 ****************************************************************************/

static int __init w9968cf_vpp_module_init(void)
{
	printk(__FILE__ ": Registering video post-processing functions...");

	inter_module_register("w9968cf_init_decoder", THIS_MODULE,
	                      w9968cf_init_decoder);
	inter_module_register("w9968cf_check_headers", THIS_MODULE,
	                      w9968cf_check_headers);
	inter_module_register("w9968cf_decode", THIS_MODULE,
	                      w9968cf_decode);
	inter_module_register("w9968cf_swap_yuvbytes", THIS_MODULE,
	                      w9968cf_swap_yuvbytes);
	inter_module_register("w9968cf_uyvy_to_rgbx", THIS_MODULE,
	                      w9968cf_uyvy_to_rgbx);
	inter_module_register("w9968cf_scale_up", THIS_MODULE,
	                      w9968cf_scale_up);

	printk("ok\n");

	return 0;
}


static void __exit w9968cf_vpp_module_exit(void)
{
	printk(__FILE__ ": Unregistering video post-processing functions...");

	inter_module_unregister("w9968cf_init_decoder");
	inter_module_unregister("w9968cf_check_headers");
	inter_module_unregister("w9968cf_decode");
	inter_module_unregister("w9968cf_swap_yuvbytes");
	inter_module_unregister("w9968cf_uyvy_to_rgbx");
	inter_module_unregister("w9968cf_scale_up");

	printk("ok\n");
}


module_init(w9968cf_vpp_module_init);
module_exit(w9968cf_vpp_module_exit);
