//---------------------------------------------------------
//	Compressed Loop
//
//		(C)2004 NAKADA
//---------------------------------------------------------

/*
 *  cloop file looks like:
 *  [32-bit uncompressed block size: network order]
 *  [32-bit number of blocks (n_blocks): network order]
 *  [64-bit file offsets of start of blocks: network order]
 *    * (n_blocks + 1).
 *  n_blocks of:
 *   [compressed block]
 *
 * Redistribution of this file is permitted under the GNU Public License.
 */

#define CLOOP_NAME	"cloop"
#define CLOOP_VERSION	"2.01"
#define CLOOP_MAX	8

/* Define this if you are using Red Hat Linux */
/* #define REDHAT_KERNEL */

#include <linux/module.h>
#include <linux/init.h>

//#include <linux/sched.h>
//#include <linux/fs.h>
#include <linux/file.h>
//#include <linux/stat.h>
//#include <linux/errno.h>
//#include <linux/major.h>
#include <linux/vmalloc.h>
//#include <linux/slab.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/blkdev.h>
//#include <asm/semaphore.h>
//#include <asm/div64.h>		/* do_div() for 64bit division */
//#include <asm/uaccess.h>
#include <linux/zutil.h>		/* Use zlib_inflate from lib/zlib_inflate */
#include <linux/loop.h>
#include "compressed_loop.h"


#ifndef MIN
#define MIN(x,y) ((x) < (y) ? (x) : (y))
#endif

#ifndef MAX
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#endif

/* Use experimental major for now */
#define MAJOR_NR		240

#define DEVICE_NAME		CLOOP_NAME
#define DEVICE_NR(device)	(MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)
#define DEVICE_NO_RANDOM
#define TIMEOUT_VALUE		(6 * HZ)

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, x...)
#endif


/* One file can be opened at module insertion time */
/* insmod cloop file=/path/to/file */
static char *file=NULL;
MODULE_PARM(file, "s");
static struct file *initial_file=NULL;


struct cloop_device
{
	/* Copied straight from the file */
	struct cloop_head head;

	/* An array of offsets of compressed blocks within the file */
	loff_t *offsets;

	/* We buffer one uncompressed `block' */
	int buffered_blocknum;
	void *buffer;
	void *compressed_buffer;

	z_stream zstream;

	struct file   *backing_file;	/* associated file */
	struct inode  *backing_inode;	/* for bmap */

	unsigned int underlying_blksize;
	int refcnt;			// Reference Count
	int isblkdev;
	struct semaphore clo_lock;

	int num;			// /dev/cloop/? cloop_num
	request_queue_t queue;
};

static int cloop_sizes[CLOOP_MAX];
static int cloop_blksizes[CLOOP_MAX];

static struct cloop_device cloop_dev[CLOOP_MAX];
static char *cloop_name=CLOOP_NAME;
static const int max_cloop = CLOOP_MAX;

static struct gendisk **disc;


#if (!(defined(CONFIG_ZLIB_INFLATE) || defined(CONFIG_ZLIB_INFLATE_MODULE))) /* Must be compiled into kernel. */
#error  "Invalid Kernel configuration. CONFIG_ZLIB_INFLATE support is needed for cloop."
#endif


/* Most of this code can be found in fs/cramfs/uncompress.c */
static int uncompress(struct cloop_device *clo, char *dest, unsigned long *dlen, char *src, unsigned long slen)
{
	int err;

	clo->zstream.next_in = src;
	clo->zstream.avail_in = slen;

	clo->zstream.next_out = dest;
	clo->zstream.avail_out = *dlen;

	err = zlib_inflateReset(&clo->zstream);
	if (err != Z_OK) {
		printk(KERN_ERR "%s: zlib_inflateReset error %d\n", cloop_name, err);
		zlib_inflateEnd(&clo->zstream);
		zlib_inflateInit(&clo->zstream);
	}

	err = zlib_inflate(&clo->zstream, Z_FINISH);
	*dlen = clo->zstream.total_out;

	if (err != Z_STREAM_END) return err;
	return Z_OK;
}


/* This is more complicated than it looks. */
struct clo_read_data
{
	struct cloop_device *clo;
	char *data; /* We need to keep track of where we are in the buffer */
	int bsize;
};


/* We need this for do_generic_file_read() because the default function */
/* wants to read into user-space for an unknown reason. :-/ See loop.c. */
// from lo_read_actor() in loop.c
static int clo_read_actor(read_descriptor_t * desc, struct page *page, unsigned long offset, unsigned long size)
{
	char *kaddr;
	struct clo_read_data *p = (struct clo_read_data*)desc->buf;
	unsigned long count = desc->count;

	if (size > count) size = count;

	kaddr = kmap(page);
	memcpy(p->data, kaddr + offset, size);
	kunmap(page);

	desc->count = count - size;
	desc->written += size;
	p->data += size;

	return size;
}


static size_t clo_read_from_file(struct cloop_device *clo, struct file *f, char *buf, loff_t pos, size_t buf_len)
{
	size_t buf_done=0;
	while (buf_done < buf_len) {
		size_t size = buf_len - buf_done;
		struct clo_read_data cd={		/* do_generic_file_read() needs this. */
			clo,				/* struct cloop_device *clo */
			(char *)(buf + buf_done),	/* char *data */
			size};				/* Actual data size */
		read_descriptor_t desc;
		desc.written = 0;
		desc.count   = size;
		desc.buf     = (char*)&cd;
		desc.error   = 0;
#ifdef REDHAT_KERNEL
		/* Red Hat Linux */
		do_generic_file_read(f, &pos, &desc, clo_read_actor, 0);
#else
		/* Normal Kernel */
		do_generic_file_read(f, &pos, &desc, clo_read_actor);
#endif
		if (desc.error||desc.written<=0) {
			int left = size - desc.written;
			if (left<0) left = 0; /* better safe than sorry */
			printk(KERN_ERR "%s: Read error at pos %Lu in file %s, %d bytes lost.\n",
				cloop_name, pos, file, left);
			memset(buf + buf_len - left, 0, left);
			break;
		}
		buf_done+=desc.written;
	}

	return buf_done;

	/*struct lo_read_data cookie;
	struct file *file;
	int retval;

	cookie.lo = lo;
	cookie.data = kmap(bvec->bv_page) + bvec->bv_offset;
	cookie.bsize = bsize;
	file = lo->lo_backing_file;
	retval = file->f_op->sendfile(file, &pos, bvec->bv_len,
			lo_read_actor, &cookie);
	kunmap(bvec->bv_page);
	return (retval < 0)? retval: 0;*/
}

/* This looks more complicated than it is */
static int load_buffer(struct cloop_device *clo, int blocknum)
{
	unsigned int buf_done = 0;
	unsigned long buflen;
	unsigned int buf_length;
	int ret;

	if (blocknum > ntohl(clo->head.num_blocks) || blocknum < 0) {
		printk(KERN_WARNING "%s: Invalid block number %d requested.\n", cloop_name, blocknum);
		clo->buffered_blocknum = -1;
		return 0;
	}

	// Read the block before ?
	if (blocknum == clo->buffered_blocknum) return 1;

	/* Is there a ntohl for 64-bit values? */
	buf_length = be64_to_cpu(clo->offsets[blocknum+1]) - be64_to_cpu(clo->offsets[blocknum]);

	/* Load one compressed block from the file. */
	clo_read_from_file(clo, clo->backing_file, (char *)clo->compressed_buffer,
			be64_to_cpu(clo->offsets[blocknum]), buf_length);

	/* Do decompression into real buffer. */
	buflen = ntohl(clo->head.block_size);

	/* Do the uncompression */
	ret = uncompress(clo, clo->buffer, &buflen, clo->compressed_buffer, buf_length);
	/* DEBUGP("cloop: buflen after uncompress: %ld\n",buflen); */
	if (ret != 0) {
		printk(KERN_ERR "%s: error %i uncompressing block %u %u/%lu/%u/%u "
			"%Lu-%Lu\n", cloop_name, ret, blocknum,
			ntohl(clo->head.block_size), buflen, buf_length, buf_done,
			be64_to_cpu(clo->offsets[blocknum]), be64_to_cpu(clo->offsets[blocknum+1]));
		clo->buffered_blocknum = -1;
		return 0;
	}
	/*printk("load_buffer: block %u %u/%lu/%u/%u %Lu-%Lu\n", blocknum,
		ntohl(clo->head.block_size), buflen, buf_length, buf_done,
		be64_to_cpu(clo->offsets[blocknum]), be64_to_cpu(clo->offsets[blocknum+1]));
	{
		int i;
		unsigned char *p;
		p=(unsigned char*)clo->buffer;
		//p=(unsigned char*)clo->compressed_buffer;
		for (i=0; i<1024; i++) printk("%x ", p[i]);
		printk("\n");
	}*/

	// Read block #
	clo->buffered_blocknum = blocknum;

	return 1;
}


// from loop_make_request() and do_lo_send() in loop.c
static int make_clo_request(request_queue_t *q, struct bio *bio)
{
	struct cloop_device *cloop;
	int status = 0;
	unsigned int len;
	loff_t offset;
	char *dest;

	int rw = bio_rw(bio);

	/* quick sanity checks */
	if (rw != READ && rw != READA) {
		DEBUGP("do_clo_request: bad command\n");
		goto out;
	}

	cloop = q->queuedata;
	if (!cloop->backing_file) {
		DEBUGP("do_clo_request: not connected to a file\n");
		goto out;
	}

	/*if (bio->bi_sector == -1) {
		DEBUGP("do_clo_request: bad sector requested\n");
		goto out;
	}*/

	unsigned vecnr;

	for (vecnr=0; vecnr<bio->bi_vcnt; vecnr++) {
		struct bio_vec *bvec = &bio->bi_io_vec[vecnr];

		len = bvec->bv_len;
		//offset = pos & ((pgoff_t)PAGE_CACHE_SIZE - 1);
		offset = ((loff_t)bio->bi_sector << 9) + bvec->bv_len * vecnr;
		dest = kmap(bvec->bv_page) + bvec->bv_offset;
		//-printk("do_clo_request: <sec:%d len:%d dest:%x>\n", (int)offset, len, (int)dest);

	down(&cloop->clo_lock);

	while (len > 0) {
		u_int32_t length_in_buffer;
		loff_t block_offset=offset;

		/* do_div (div64.h) returns the 64bit division remainder and  */
		/* puts the result in the first argument, i.e. block_offset   */
		/* becomes the blocknumber to load, and offset_in_buffer the  */
		/* position in the buffer */
		u_int32_t offset_in_buffer;
		offset_in_buffer = do_div(block_offset, ntohl(cloop->head.block_size));

		status=load_buffer(cloop, block_offset);
		//-printk("do_clo_request: %d\n", status);
		if (!status) break; /* invalid data, leave inner loop, goto next request */

		/* Now, at least part of what we want will be in the buffer. */
		length_in_buffer = ntohl(cloop->head.block_size) - offset_in_buffer;

		if (length_in_buffer > len) {
			/*     DEBUGP("Warning: length_in_buffer=%u > len=%u\n",
						length_in_buffer,len); */
			length_in_buffer = len;
		}

		memcpy(dest, cloop->buffer + offset_in_buffer, length_in_buffer);
		//-printk("do_clo_request: %d\n", length_in_buffer);

		dest   += length_in_buffer;
		len    -= length_in_buffer;
		offset += length_in_buffer;
	} /* while inner loop */

	up(&cloop->clo_lock);

	}

out:
	bio_endio(bio, bio->bi_size, 0);
	//-bio_io_error(bio, bio->bi_size);
	return 0;
}

/* Read header and offsets from already opened file */
// from loop_set_fd() in loop.c
static int clo_set_file(struct cloop_device *clo, struct file *file, char *filename)
{
	struct inode *inode;
	char *bbuf=NULL;
	unsigned int i, offsets_read, total_offsets;
	unsigned long largest_block=0;
	int isblkdev;
	int error = 0;

	/* This is safe, since we have a reference from open(). */
	__module_get(THIS_MODULE);

	inode = file->f_dentry->d_inode;
	isblkdev=S_ISBLK(inode->i_mode)?1:0;

	if (!isblkdev && !S_ISREG(inode->i_mode)) {
		printk(KERN_ERR "%s: %s not a regular file or block device\n", cloop_name, filename);
		error=-EBADF;
		goto error_release;
	}

	clo->backing_file = file;
	clo->backing_inode= inode;

	if (!isblkdev && inode->i_size < sizeof(struct cloop_head)) {
		printk(KERN_ERR "%s: %lu bytes (must be >= %u bytes)\n",
			cloop_name, (unsigned long)inode->i_size, (unsigned)sizeof(struct cloop_head));
		error=-EBADF;
		goto error_release;
	}

	/* Get initial block size out of device */
	clo->underlying_blksize = isblkdev ? block_size(inode->i_bdev) : inode->i_blksize;
	DEBUGP("Underlying blocksize is %u\n", clo->underlying_blksize);

	bbuf = vmalloc(clo->underlying_blksize);
	if (!bbuf) {
		printk(KERN_ERR "%s: out of kernel mem for block buffer (%lu bytes)\n",
			cloop_name, (unsigned long)clo->underlying_blksize);
		error=-ENOMEM;
		goto error_release;
	}

	total_offsets = 1; /* Dummy total_offsets: will be filled in first time around */
	for (i=0, offsets_read=0; offsets_read<total_offsets; i++) {
		unsigned int offset = 0, num_readable;

		/* Kernel 2.4 version */
		size_t bytes_read = clo_read_from_file(clo, file, bbuf,
					i*clo->underlying_blksize, clo->underlying_blksize);
		if (bytes_read != clo->underlying_blksize) { error=-EBADF; goto error_release; }

		/* Header will be in block zero */
		if (i==0) {
			memcpy(&clo->head, bbuf, sizeof(struct cloop_head));
			offset = sizeof(struct cloop_head);
			if (ntohl(clo->head.block_size) % 512 != 0) {
				printk(KERN_ERR "%s: blocksize %u not multiple of 512\n",
					cloop_name, ntohl(clo->head.block_size));
				error=-EBADF; goto error_release;
			}

			if (clo->head.preamble[0x0B]!='V'||clo->head.preamble[0x0C]<'1') {
				printk(KERN_ERR "%s: Cannot read old 32-bit (version 0.68) images, "
						"please use an older version of %s for this file.\n",
						cloop_name, cloop_name);
				error=-EBADF; goto error_release;
			}

			if (clo->head.preamble[0x0C]<'2') {
				printk(KERN_ERR "%s: Cannot read old architecture-dependent "
						"(format <= 1.0) images, please use an older "
						"version of %s for this file.\n",
						cloop_name, cloop_name);
				error=-EBADF; goto error_release;
			}

			total_offsets=ntohl(clo->head.num_blocks)+1;

			if (!isblkdev && (sizeof(struct cloop_head)+sizeof(loff_t)*total_offsets > inode->i_size)) {
				printk(KERN_ERR "%s: file too small for %u blocks\n",
					cloop_name, ntohl(clo->head.num_blocks));
				error=-EBADF; goto error_release;
			}

			clo->offsets = vmalloc(sizeof(loff_t) * total_offsets);
			if (!clo->offsets) {
				printk(KERN_ERR "%s: out of kernel mem for offsets\n", cloop_name);
				error=-ENOMEM; goto error_release;
			}
		}

		num_readable = MIN(total_offsets - offsets_read,
				(clo->underlying_blksize - offset) / sizeof(loff_t));
		memcpy(&clo->offsets[offsets_read], bbuf+offset, num_readable * sizeof(loff_t));
		offsets_read += num_readable;
	}

	{ /* Search for largest block rather than estimate. */
	int i;
	for (i=0; i<total_offsets-1; i++) {
		loff_t d=be64_to_cpu(clo->offsets[i+1]) - be64_to_cpu(clo->offsets[i]);
		largest_block=MAX(largest_block, d);
	}
	printk("%s: %s: %u blocks, %u bytes/block, largest block is %lu bytes.\n",
		cloop_name, filename, ntohl(clo->head.num_blocks), ntohl(clo->head.block_size), largest_block);
	}

	/* Combo kmalloc used too large chunks (>130000). */
	clo->buffer = vmalloc(ntohl(clo->head.block_size));
	if (!clo->buffer) {
		printk(KERN_ERR "%s: out of memory for buffer %lu\n",
			cloop_name, (unsigned long)ntohl(clo->head.block_size));
		error=-ENOMEM;
		goto error_release_free;
	}

	clo->compressed_buffer = vmalloc(largest_block);

	if (!clo->compressed_buffer) {
		printk(KERN_ERR "%s: out of memory for compressed buffer %lu\n",
			cloop_name, largest_block);
		error=-ENOMEM; goto error_release_free_buffer;
	}

	// from cramfs_uncompress_init()
	clo->zstream.workspace = vmalloc(zlib_inflate_workspacesize());
	if (!clo->zstream.workspace) {
		printk(KERN_ERR "%s: out of mem for zlib working area %u\n",
			cloop_name, zlib_inflate_workspacesize());
		error=-ENOMEM;
		goto error_release_free_all;
	}
	zlib_inflateInit(&clo->zstream);

	if (!isblkdev && be64_to_cpu(clo->offsets[ntohl(clo->head.num_blocks)]) != inode->i_size) {
		printk(KERN_ERR "%s: final offset wrong (%Lu not %Lu)\n",
			cloop_name,
			be64_to_cpu(clo->offsets[ntohl(clo->head.num_blocks)]),
			inode->i_size);
		vfree(clo->zstream.workspace); clo->zstream.workspace=NULL;
		goto error_release_free_all;
	}

	clo->buffered_blocknum = -1;
	cloop_sizes[clo->num] = ntohl(clo->head.num_blocks)
			* ( ntohl(clo->head.block_size) / BLOCK_SIZE );
	/* this seems to be the maximum allowed blocksize (Kernel limit) */
	cloop_blksizes[clo->num] = PAGE_SIZE;

	/*
	 * set queue make_request_fn, and add limits based on lower level
	 * device
	 */
	blk_queue_make_request(&clo->queue, make_clo_request);
	clo->queue.queuedata = clo;

	return error;

error_release_free_all:
	vfree(clo->compressed_buffer);
	clo->compressed_buffer=NULL;

error_release_free_buffer:
	vfree(clo->buffer);
	clo->buffer=NULL;

error_release_free:
	vfree(clo->offsets);
	clo->offsets=NULL;

error_release:
	if (bbuf) vfree(bbuf);
	clo->backing_file=NULL;

	/* This is safe: open() is still holding a reference. */
	module_put(THIS_MODULE);

	return error;
}

/* Code adapted from Theodore Ts'o's linux/drivers/block/loop.c */
/* Get file from ioctl arg (losetup) */
static int clo_set_fd(struct cloop_device *clo, struct file *clo_file, struct block_device *bdev, unsigned int arg)
{
	struct file *file=NULL;
	int error = 0;

	/* Already an allocated file present */
	if (clo->backing_file) return -EBUSY;

	file = fget(arg); /* get filp struct from ioctl arg fd */
	if (!file) return -EBADF;

	error=clo_set_file(clo, file, "losetup_file");
	if (error) fput(file);

	return error;
}


static int clo_clr_fd(struct cloop_device *clo, struct block_device *bdev)
{
	struct file *filp = clo->backing_file;

	/* we needed one fd for the ioctl */
	if (clo->refcnt > 1) return -EBUSY;
	if (filp==NULL) return -EINVAL;
	if (filp!=initial_file) {
		fput(filp);
	} else {
		filp_close(initial_file, 0);
		initial_file=NULL;
	}

	clo->backing_file  = NULL;
	clo->backing_inode = NULL;
	cloop_sizes[clo->num] = 0;
	cloop_blksizes[clo->num] = 0;

	/* This is safe: open() is still holding a reference. */
	module_put(THIS_MODULE);

	return 0;
}

static int clo_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct cloop_device *clo = inode->i_bdev->bd_disk->private_data;
	int err;

	/* We use the same ioctls that loop does */
	switch (cmd) {
	case LOOP_SET_FD:
		err = clo_set_fd(clo, file, inode->i_bdev, arg);
		break;
	case LOOP_CLR_FD:
		err = clo_clr_fd(clo, inode->i_bdev);
		break;
	case LOOP_SET_STATUS:
	case LOOP_GET_STATUS:
		err=0;
		break;
	default:
		err = -EINVAL;
	}

	return err;
}


// from lo_open() in loop.c
static int clo_open(struct inode *inode, struct file *file)
{
	struct cloop_device *clo = inode->i_bdev->bd_disk->private_data;
	clo->refcnt++;
	//-printk("%s: <refcnt %d>\n", cloop_name, clo->refcnt);

	return 0;
}


// from lo_release() in loop.c
static int clo_close(struct inode *inode, struct file *file)
{
	struct cloop_device *clo = inode->i_bdev->bd_disk->private_data;
	clo->refcnt--;
	//-printk("%s: <refcnt %d>\n", cloop_name, clo->refcnt);

	return 0;
}


static struct block_device_operations clo_fops = {
	.owner =	THIS_MODULE,
	.open =		clo_open,
	.release =	clo_close,
	.ioctl =	clo_ioctl,
};


// from loop_init() in loop.c
int __init init_cloop_fs(void)
{
	int i, error=0;
	printk("%s: Initializing cloop ver "CLOOP_VERSION"\n", cloop_name);

	if (register_blkdev(MAJOR_NR, cloop_name)) {
		printk(KERN_WARNING "%s: Unable to get major %d for cloop\n", cloop_name, MAJOR_NR);
		if (initial_file) { error=-EIO; goto error_filp_close; }
	}

	disc = kmalloc(max_cloop * sizeof(struct gendisk *), GFP_KERNEL);
	//if (!disc) goto out_mem2;
	for (i=0; i<max_cloop; i++) {
		disc[i] = alloc_disk(1);
		//if (!disks[i]) goto out_mem3;
	}

	devfs_mk_dir("cloop");
	for (i=0; i<max_cloop; i++) {
		memset(&cloop_dev[i], 0, sizeof(struct cloop_device));
		init_MUTEX(&cloop_dev[i].clo_lock);
		cloop_dev[i].num=i;

		disc[i]->major = MAJOR_NR;
		disc[i]->first_minor = i;
		disc[i]->fops = &clo_fops;
		sprintf(disc[i]->disk_name, "cloop%d", i);
		sprintf(disc[i]->devfs_name, "cloop/%d", i);
		disc[i]->private_data = &cloop_dev[i];
		disc[i]->queue = &cloop_dev[i].queue;
		add_disk(disc[i]);
	}

	/* global file name for first cloop-device is a module option string. */
	if (file) {
		initial_file=filp_open(file, 0x00, 0x00);
		if (initial_file==NULL || IS_ERR(initial_file)) {
			printk(KERN_ERR "%s: Unable to get file %s for cloop device\n", cloop_name, file);
			return -EINVAL;
		}
		error=clo_set_file(&cloop_dev[0], initial_file, file);
		if (error) goto error_filp_close;
	}

	printk(KERN_INFO "cloop: loaded (max %d devices)\n", max_cloop);

	return 0;

error_filp_close:
	//unregister_blkdev(MAJOR_NR, cloop_name);
	filp_close(initial_file, 0);
	initial_file=NULL;
	cloop_dev[0].backing_file=NULL;
	return error;
}


// from loop_exit() in loop.c
void exit_cloop_fs(void)
{
	int i;

	for (i=0; i<max_cloop; i++) {
		del_gendisk(disc[i]);
		blk_put_queue(&cloop_dev[i].queue);
		put_disk(disc[i]);
	}

	devfs_remove("cloop");
	if (unregister_blkdev(MAJOR_NR, cloop_name) != 0) {
		printk(KERN_WARNING "%s: cannot unregister block device\n", cloop_name);
	}

	for (i=0; i<max_cloop; i++) {
		if (cloop_dev[i].offsets) vfree(cloop_dev[i].offsets);
		if (cloop_dev[i].buffer) vfree(cloop_dev[i].buffer);
		if (cloop_dev[i].compressed_buffer) vfree(cloop_dev[i].compressed_buffer);

		// from cramfs_uncompress_exit()
		zlib_inflateEnd(&cloop_dev[i].zstream);
		if (cloop_dev[i].zstream.workspace) vfree(cloop_dev[i].zstream.workspace);

		if (cloop_dev[i].backing_file && cloop_dev[i].backing_file!=initial_file) {
			fput(cloop_dev[i].backing_file);
		}
	}

	if (initial_file) filp_close(initial_file, 0);
	printk("%s: unloaded.\n", cloop_name);

	return;
}


module_init(init_cloop_fs);
module_exit(exit_cloop_fs);
MODULE_LICENSE("GPL");
