/*******************************************************************************
  
  Copyright(c) 2003 - 2004 Intel Corporation. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information:
  James P. Ketrenos <ipw2100-admin@linux.intel.com>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/
#ifndef _IPW2100_FRAG_H
#define _IPW2100_FRAG_H

#define IS_ESS(priv)     (priv->port_type == BSS)
#define IS_IBSS(priv)    (priv->port_type == IBSS)

enum {
        WSE_UNICAST,
        WSE_MULTICAST,
        WSE_BROADCAST
};

#define ETH_IS_BROADCAST(addr)  (addr[0] == 0xff && addr[1] == 0xff     \
                                 && addr[2] == 0xff && addr[3] == 0xff  \
                                 && addr[4] == 0xff && addr[5] == 0xff)

#define ETH_IS_MULTICAST(addr)  (addr[0] & 0x01)

#define FRAGMENT_NUMBER_FIELD_MASK				0x000F
#define SEQUENCE_NUMBER_FIELD_MASK				0xFFF0
#define SEQUENCE_NUMBER_FIELD_SHIFT				4

#define FRAGMENT_NUMBER(header) ((header->seq_ctrl & \
        FRAGMENT_NUMBER_FIELD_MASK))
#define SEQUENCE_NUMBER(header) ((header->seq_ctrl & \
        SEQUENCE_NUMBER_FIELD_MASK)>>SEQUENCE_NUMBER_FIELD_SHIFT)

#define IS_FRAGMENT(header) (((*header).flags == 1) || \
                            (FRAGMENT_NUMBER(header) > 0))

#define DEFAULT_FRAG_THRESHOLD 2342U
#define MIN_FRAG_THRESHOLD     256U
#define	MAX_FRAG_THRESHOLD     2342U

#define MAX_DEFRAG_TIMEOUT	0xFFFFFFFF	
#define MIN_DEFRAG_TIMEOUT	0x1
#define DEFAULT_DEFRAG_TIMEOUT	0x200

struct sk_buff *ipw2100_defrag_rx(struct ipw2100_priv *priv, 
				  struct ipw2100_hw_packet *packet, 
				  u32 frame_size);
int ipw2100_defrag_alloc(struct ipw2100_priv *);
void ipw2100_defrag_free(struct ipw2100_priv *);
void ipw2100_defrag_initialize(struct ipw2100_priv *priv);

u32 ipw2100_msdu_frag_threshold(struct ipw2100_priv *);

#endif
