// ****************************************************************************
//
//              CGina.cpp
//
//              Implementation file for the CGina driver class.  CGina is for 
//              20-bit Gina, not Gina24.
//
//              Set editor tabs to 3 for your viewing pleasure.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//              
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//              
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//              
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//              
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHOGALS_FAMILY
#include "Echoaudio.h"




struct CEchoGals *NewEchoGals(u16 Device, u16 SubDevice) {
  struct CEchoGals *eg;
  int i;

ECHO_DEBUGPRINTF(("NewEchoGals - Gina20\n"));
  if ((SubDevice&0xfff0)!=0x0020) {
    ECHO_DEBUGPRINTF(("NewEchoGals() - wrong card type\n"));
    return(0);
  }
  eg=OsAllocateNonPaged(sizeof(struct CEchoGals));
  if (eg) {
    memset(eg, 0, sizeof(struct CEchoGals));
  } else {
    ECHO_DEBUGPRINTF(("NewEchoGals() - memory allocation failed\n"));
    return(0);
  }

  eg->DeviceId=Device;
  eg->SubDeviceId=SubDevice;
  eg->CardType=0x0020;
  eg->BadBoard=1;
  eg->Flags = ECHOGALS_FLAG_BADBOARD | ECHOGALS_FLAG_SPDIF_NODITHER;
  for (i=0; i<ECHO_MAXAUDIOPIPES; i++)
    eg->Pipes[i].BytesPerSample=1;
  return(eg);
}



int InitHw(struct CEchoGals *eg) {
  int Status;

  ECHO_DEBUGPRINTF(("InitHw()\n"));
  // Create the DSP comm object
  Status = NewDspComm(eg);
  if (Status) {
    ECHO_DEBUGPRINTF(("InitHw - could not create DSP comm object\n"));
    return(ECHOSTATUS_NO_MEM);
  }

  // Load the DSP and the ASIC on the PCI card
  Status = LoadFirmware(eg);
  if (Status || eg->BadBoard)
    return(ECHOSTATUS_DSP_DEAD);

  // Clear the "bad board" flag
  eg->Flags &= ~ECHOGALS_FLAG_BADBOARD;

  // Must call this here after DSP is init to init gains and mutes
  Status = InitLineLevels(eg);
  if (ECHOSTATUS_OK != Status)
    return(Status);

  // Set the S/PDIF output format to "professional"
  SetProfessionalSpdif(eg, TRUE);

  // Get default sample rate from DSP
  eg->SampleRate = GetSampleRate(eg);

  ECHO_DEBUGPRINTF(("InitHw done\n"));
  return(Status);
}



int GetCapabilities(struct CEchoGals *eg, PECHOGALS_CAPS pCapabilities) {
  int Status, i;

  Status = GetBaseCapabilities(eg, pCapabilities);

  if (ECHOSTATUS_OK != Status)
    return(Status);

  // Add input gain to input busses
  for (i=0; i<eg->FirstDigitalBusIn; i++)
    pCapabilities->dwBusInCaps[i] |= ECHOCAPS_GAIN;

  pCapabilities->dwInClockTypes |= ECHO_CLOCK_BIT_SPDIF;
  return(Status);
}



//===========================================================================
//
// QueryAudioSampleRate is used to find out if this card can handle a given sample rate.
//
//===========================================================================

int QueryAudioSampleRate(struct CEchoGals *eg, u32 SampleRate) {

  if (SampleRate!=44100 && SampleRate!=48000) {
    ECHO_DEBUGPRINTF(("QueryAudioSampleRate() Sample rate must be 44,100 Hz or 48,000 Hz\n"));
    return(ECHOSTATUS_BAD_FORMAT);
  }

  ECHO_DEBUGPRINTF(("QueryAudioSampleRate()\n"));
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// GetInputClockDetect returns a bitmask consisting of all the input
// clocks currently connected to the hardware; this changes as the user
// connects and disconnects clock inputs.
//
// You should use this information to determine which clocks the user is
// allowed to select.
//
// Gina20 only supports S/PDIF input clock.
//
//===========================================================================

u32 GetInputClockDetect(const struct CEchoGals *eg) {
  u32 ClocksFromDsp, ClockDetectBits;

  ECHO_DEBUGPRINTF(("GetInputClockDetect\n"));

  // Map the DSP clock detect bits to the generic driver clock detect bits
  ClocksFromDsp = DspGetInputClockDetect(eg);

  ClockDetectBits = ECHO_CLOCK_BIT_INTERNAL;

  if (ClocksFromDsp & GLDM_CLOCK_DETECT_BIT_SPDIF)
    ClockDetectBits |= ECHO_CLOCK_BIT_SPDIF;

  return(ClockDetectBits);
}

