// ****************************************************************************
//
//              CDaffyDuck.CPP
//
//              Implementation file for the CDaffyDuck class.
//              Set editor tabs to 3 for your viewing pleasure.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//              
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//              
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//              
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
//---------------------------------------------------------------------------
//
// The head pointer tracks the next free slot in the circular buffers
// The tail pointer tracks the oldest mapping.
//
//****************************************************************************

#define ECHO24_FAMILY                   // Not used, only for compile
#include "Echoaudio.h"



int AllocateSGList(struct CEchoGals *eg, u16 PipeIndex) {

  // Allocate the page for the scatter-gather list
  if (ECHOSTATUS_OK != OsPageAllocate(eg->ossprivate_data, 1, (void **)&eg->Pipes[PipeIndex].sglist, (dma_addr_t *)&eg->Pipes[PipeIndex].sglist_phys)) {
    ECHO_DEBUGPRINTF(("SGlist: memory allocation failed\n"));
    return(ECHOSTATUS_NO_MEM);
  }
  memset(eg->Pipes[PipeIndex].sglist, 0, PAGE_SIZE);
  eg->Pipes[PipeIndex].sgHead=0;
  eg->DspCommPage->SGListPhys[PipeIndex].PhysAddr = SWAP32(eg->Pipes[PipeIndex].sglist_phys);
  ECHO_DEBUGPRINTF(("SGlist: memory allocated phys=%x\n", eg->Pipes[PipeIndex].sglist_phys));
  return(ECHOSTATUS_OK);
}



void FreeSGList(struct CEchoGals *eg, u16 PipeIndex) {

  ECHO_DEBUGPRINTF(("SGlist: free phys=%x\n", eg->Pipes[PipeIndex].sglist_phys));
  OsPageFree(eg->ossprivate_data, 1, eg->Pipes[PipeIndex].sglist, eg->Pipes[PipeIndex].sglist_phys);
  eg->Pipes[PipeIndex].sglist=0;
  eg->Pipes[PipeIndex].sglist_phys=0;
}



int SGListAddMapping(struct CEchoGals *eg, u16 PipeIndex, dma_addr_t Address, size_t Length, char Irq) {
int head;

  head=eg->Pipes[PipeIndex].sgHead;
//  ECHO_DEBUGPRINTF(("SGlist: add[%2d]  phys=%x  len=%d\n", head, Address, Length));
  if (head<MAX_SGLIST_ENTRIES-1) {
    eg->Pipes[PipeIndex].sglist[head].PhysAddr=SWAP32(Address);
    eg->Pipes[PipeIndex].sglist[head].Size=SWAP32(Length);
    head++;
  } else {
    ECHO_DEBUGPRINTF(("SGlist: too many fragments\n"));
    return(ECHOSTATUS_DUCK_FULL);
  }

  if (Irq) {
    if (head<MAX_SGLIST_ENTRIES-1) {
      eg->Pipes[PipeIndex].sglist[head].PhysAddr=0;
      eg->Pipes[PipeIndex].sglist[head].Size=0;
      head++;
    } else {
      ECHO_DEBUGPRINTF(("SGlist: too many fragments\n"));
      return(ECHOSTATUS_DUCK_FULL);
    }
  }

  eg->Pipes[PipeIndex].sgHead=head;
  return(ECHOSTATUS_OK);
}



int SGListWrap(struct CEchoGals *eg, u16 PipeIndex) {

  return(SGListAddMapping(eg, PipeIndex, eg->Pipes[PipeIndex].sglist_phys, 0, 0));
}

