// ****************************************************************************
//
//		OsSupportWDM.H
//
//		Include file for WDM Support Services to the CEchoGals
//		generic driver class
//
//		Set editor tabs to 3 for your viewing pleasure.
//
// ----------------------------------------------------------------------------
//
//		Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//		All rights reserved
//		www.echoaudio.com
//
//		Permission is hereby granted, free of charge, to any person obtaining a
//		copy of this software and associated documentation files (the
//		"Software"), to deal with the Software without restriction, including
//		without limitation the rights to use, copy, modify, merge, publish,
//		distribute, sublicense, and/or sell copies of the Software, and to
//		permit persons to whom the Software is furnished to do so, subject to
//		the following conditions:
//
//		- Redistributions of source code must retain the above copyright
//		notice, this list of conditions and the following disclaimers.
//
//		- Redistributions in binary form must reproduce the above copyright
//		notice, this list of conditions and the following disclaimers in the
//		documentation and/or other materials provided with the distribution.
//
//		- Neither the name of Echo Digital Audio, nor the names of its
//		contributors may be used to endorse or promote products derived from
//		this Software without specific prior written permission.
//
//		THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//		EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//		MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//		IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//		ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//		TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//		SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************


// Prevent problems with multiple includes
#ifndef _ECHOOSSUPPORT_
#define _ECHOOSSUPPORT_

#include "OsSupport_types.h"



#if (CONFIG_SND_DEBUG_FULL)
#define ECHO_DEBUGPRINTF(strings) snd_printk strings
#else
#define ECHO_DEBUGPRINTF(strings)
#endif



// Define generic byte swapping functions

#ifdef __BIG_ENDIAN
#define SWAP16(x) le16_to_cpu(x)
#define SWAP32(x) le32_to_cpu(x)
#else
#define SWAP16(x) (x)
#define SWAP32(x) (x)
#endif




//#define _DEBUG


/* Memory barriers

 MB prevents loads and stores being reordered across this point.
 RMB prevents loads being reordered across this point.
 WMB prevents stores being reordered across this point.

 On PowerPC the eieio instruction is a barrier providing an ordering (separately) for (a) cacheable
 stores and (b) loads and stores to non-cacheable memory (e.g. I/O devices).

 For now, "wmb()" doesn't actually do anything, as all
 Intel CPU's follow what Intel calls a *Processor Order*,
 in which all writes are seen in the program order even
 outside the CPU.
 I expect future Intel CPU's to have a weaker ordering,
 but I'd also expect them to finally get their act together
 and add some real memory barriers if so.
#define MB wmb()
#define RMB wmb()
#define WMB wmb()
*/

#define MB
#define RMB
#define WMB


//
//	Global Memory Management Functions
//

/*
char *strcpy(char *, const char *);
char *strncpy(char *, const char *, size_t);
void *memset(void *, int, size_t);
void *memmove(void *, const void *, size_t);
#if __GNUC__ == 3
void *memcpy(void *, const void *, size_t);
#endif
*/



// Allocate locked, non-pageable block of memory.  Does not have to be
// physically contiguous.  Primarily used to implement the overloaded
// new operator for classes that must remain memory resident.
void *OsAllocateNonPaged(size_t Bytes);

// Unlock and free, non-pageable block of memory.
int OsFreeNonPaged(void *MemAddr);

// Copy memory
#define OsCopyMemory(pDest, pSrc, dwBytes) memcpy(pDest, pSrc, dwBytes)

// Set memory to zero
#define OsZeroMemory(pDest, dwBytes) memset(pDest, 0, dwBytes)

#ifndef memset
void *memset(void *s, int c, size_t n);
#endif

#ifndef memcpy
void *memcpy(void *dest, const void *src, size_t n);
#endif

  //
  // Timer Methods
  //

  // Return the system time in microseconds.
  // Return error status if the OS doesn't support this function.
int OsGetSystemTime(u64 *Time);


  // Stall execution for dwTime microseconds.
  // Return error status if the OS doesn't support this function.
int OsSnooze(u32 Time);


  //
  // Memory Management Methods
  //

  // Allocate locked, non-pageable, physically contiguous memory pages
  // in the drivers address space.  Used to allocate memory for the DSP
  // communications area and Ducks.
  //  EchoGals object
  //  How many pages to allocate
  //  Where to return the memory ptr
  //  Where to return the physical PCI addr
int OsPageAllocate(const void *pci, u32 Pages, void **PageAddr, dma_addr_t *PhysicalPageAddr);

  // Unlock and free non-pageable, physically contiguous memory pages
  // in the drivers address space.
  // Used to free memory for the DSP communications area and Ducks.
  //  EchoGals object
  //  How many pages to free
  //  Virtual memory ptr
  //  Physical PCI addr
int OsPageFree(const void *pci, u32 Pages, void *PageAddr, dma_addr_t PhysicalPageAddr);


#endif  // _ECHOOSSUPPORT_
